package de.unihannover.orderingExperimentAnalysis;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import de.unihannover.orderingExperimentAnalysis.SessionData.Defect;
import de.unihannover.orderingExperimentAnalysis.SessionData.FP;

public class NavigationTrace {

    private static final double TIME_FACTOR = 3000.0;
    private static final double PIXEL_FACTOR = 100.0;

    private final long startTime;
    private final String id;
    private final Patch change;
    private final Treatment treatment;
    private final int reviewNumber;
    private final int viewHeight;
    private long endTime;
    private final List<ScrollPoint> scrollPoints = new ArrayList<>();
    private final List<Interaction> interactions = new ArrayList<>();

    private static final class ScrollPoint {
        private final long timeSinceStart;
        private final long y;

        public ScrollPoint(long l, long yPos) {
            this.timeSinceStart = l;
            this.y = yPos;
        }
    }

    private static enum Color {
        REMARK_REAL("1 0 0"),
        REMARK_FP("1 0 0.5"),
        REMARK_REMOVED("0.8 0 0.2"),
        STATE_CHANGE("0 0 1"),
        CONTEXT("0.8 0.8 0"),
        MOUSE("0.2 0.8 0.2"),
        KEY("0.2 0.8 0.5");

        private String color;

        private Color(String color) {
            this.color = color;
        }
    }

    private static final class Interaction {
        private final long timeSinceStart;
        private final String msg;
        private final Color color;

        public Interaction(long l, String msg, Color color) {
            this.timeSinceStart = l;
            this.msg = msg;
            this.color = color;
        }
    }

    public NavigationTrace(String id, Patch change, Treatment treatment, int reviewNumber, long startTime,
                    int viewHeight) {
        this.id = id;
        this.change = change;
        this.treatment = treatment;
        this.reviewNumber = reviewNumber;
        this.startTime = startTime;
        this.viewHeight = viewHeight;
    }

    public void addScrollTo(long time, long yPos) {
        this.scrollPoints.add(new ScrollPoint(time - this.startTime, yPos));
    }

    public void addStateChange(long time, boolean isStart, String stateName) {
        this.interactions.add(new Interaction(time - this.startTime, (isStart ? "start " : "end ") + stateName, Color.STATE_CHANGE));
    }

    public void addReviewEnded(long time) {
        assert time > this.startTime;
        this.endTime = time;
    }

    public void addRemark(long time, String pos, Defect defect) {
        this.interactions.add(new Interaction(time - this.startTime, "+" + pos + " " + defect.getShortString(), defect == FP.FP ? Color.REMARK_FP : Color.REMARK_REAL));
    }

    public void changeRemark(long time, String pos, Defect defect) {
        this.interactions.add(new Interaction(time - this.startTime, "~" + pos + " " + defect.getShortString(), defect == FP.FP ? Color.REMARK_FP : Color.REMARK_REAL));
    }

    public void deleteRemark(long time, String pos) {
        this.interactions.add(new Interaction(time - this.startTime, "-" + pos, Color.REMARK_REMOVED));
    }

    public void addMouseClick(long time, char key, boolean alt, boolean shift, boolean ctrl, boolean meta) {
        final StringBuilder s = new StringBuilder();
        s.append("MC ");
        if (ctrl) {
            s.append("Ctrl+");
        }
        if (alt) {
            s.append("Alt+");
        }
        if (shift) {
            s.append("Shift+");
        }
        if (meta) {
            s.append("Meta+");
        }
        s.append(key == '0' ? "Le" : (key == '2' ? "Ri" : Character.toString(key)));
        this.interactions.add(new Interaction(time - this.startTime, s.toString(), Color.MOUSE));
    }

    public void addKeyPress(long time, int key, boolean alt, boolean shift, boolean ctrl) {
        final StringBuilder s = new StringBuilder();
        if (ctrl) {
            s.append("Ctrl+");
        }
        if (alt) {
            s.append("Alt+");
        }
        if (shift) {
            s.append("Shift+");
        }
        switch (key) {
        case 38:
            s.append("up");
            break;
        case 40:
            s.append("down");
            break;
        case 70:
            s.append("f");
            break;
        case 13:
            s.append("enter");
            break;
        case 112:
            s.append("F1");
            break;
        case 115:
            s.append("F4");
            break;
        case 99:
            s.append("pgdwn(num)");
            break;
        default:
            s.append("kc" + key);
            break;
        }
        this.interactions.add(new Interaction(time - this.startTime, s.toString(), Color.KEY));
    }

    private static final class PsPrinter {
        private final PrintStream out;

        public PsPrinter(PrintStream out2) {
            this.out = out2;
        }

        public void moveto(double x, double y) {
            this.out.println(x + " " + y + " moveto");
        }

        public void lineto(double x, double y) {
            this.out.println(x + " " + y + " lineto");
        }

        public void raw(String s) {
            this.out.print(s);
        }

        public void rawln(String s) {
            this.out.println(s);
        }

        public void show(String string) {
            this.out.println("(" + string + ")showflip");
        }
    }

    public void toPostscript(PrintStream out) {
        this.cleanData();

        out.println("%%PageMedia: m" + this.getPageWidth());
        out.println("%%BeginPageSetup");
        out.println("<< /PageSize [" + this.getPageWidth() + " 595] >> setpagedevice");
        out.println("%%EndPageSetup");
        out.println("gsave");
        out.println("10 500 translate 1 -1 scale");
        final PsPrinter ps = new PsPrinter(out);
        this.printTimeline(ps);
        this.printScrollPattern(ps);
        this.printInteraction(ps);
        out.println("grestore");
    }

    private void printInteraction(PsPrinter ps) {
        final int offset = ((int) (this.getMaxScroll() / PIXEL_FACTOR)) + 10;

        for (int i = 0; i < this.interactions.size(); i++) {
            final Interaction cur = this.interactions.get(i);
            ps.rawln(cur.color.color + " setrgbcolor");
            ps.moveto(cur.timeSinceStart / TIME_FACTOR, offset + 2);
            ps.lineto(cur.timeSinceStart / TIME_FACTOR, offset + 4);
            ps.rawln("stroke");
            final int y = offset + 25 + (i % 17) * 15;
            ps.moveto(cur.timeSinceStart / TIME_FACTOR, y - 1);
            ps.lineto(cur.timeSinceStart / TIME_FACTOR, y + 1);
            ps.rawln("stroke");
            ps.moveto(cur.timeSinceStart / TIME_FACTOR + 2, y);
            ps.show(cur.msg);
        }
    }

    private long getMaxScroll() {
        long ret = 0;
        for (final ScrollPoint p : this.scrollPoints) {
            ret = Math.max(ret, p.y);
        }
        return ret;
    }

    private void printTimeline(PsPrinter ps) {
        final long duration = this.endTime - this.startTime;
        ps.moveto(0, -3);
        ps.lineto(duration / TIME_FACTOR, -3);
        ps.rawln("stroke");
        for (long i = 0; i <= duration; i += 5 * 60 * 1000) {
            ps.moveto(i / TIME_FACTOR, -5);
            ps.lineto(i / TIME_FACTOR, -1);
            ps.rawln("stroke");
            ps.moveto(i / TIME_FACTOR, -6);
            ps.show(Long.toString(i / 60 / 1000));
        }
    }

    public int getPageWidth() {
        final int neededWidth = ((int) ((this.endTime - this.startTime) / TIME_FACTOR)) + 30;
        return Math.max(842, neededWidth);
    }

    private void cleanData() {
        this.scrollPoints.removeIf((ScrollPoint p) -> p.timeSinceStart > this.endTime - this.startTime);
        Collections.sort(this.scrollPoints, Comparator.comparingLong((ScrollPoint s) -> s.timeSinceStart));
        this.interactions.removeIf((Interaction p) -> p.timeSinceStart > this.endTime - this.startTime);
        Collections.sort(this.interactions, Comparator.comparingLong((Interaction s) -> s.timeSinceStart));
    }

    private void printScrollPattern(PsPrinter out) {
        out.moveto(0, -20);
        out.show(this.id + ", R" + this.reviewNumber + ", " + this.change + ", " + this.treatment);
        out.raw("newpath ");
        this.printScrollPath(out, 0);
        out.rawln("stroke");
        out.raw("newpath ");
        this.printScrollPath(out, this.viewHeight);
        out.rawln("stroke");
    }

    private void printScrollPath(PsPrinter out, long offset) {
        double prevY = offset / PIXEL_FACTOR;
        out.moveto(0, prevY);
        for (final ScrollPoint p : this.scrollPoints) {
            //paint as rectangular move
            out.lineto(p.timeSinceStart / TIME_FACTOR, prevY);
            prevY = (p.y + offset) / PIXEL_FACTOR;
            out.lineto(p.timeSinceStart / TIME_FACTOR, prevY);
        }
    }

    public void showContext(long time) {
        this.interactions.add(new Interaction(time - this.startTime, "show ctx", Color.CONTEXT));
    }

    public void hideContext(long time) {
        this.interactions.add(new Interaction(time - this.startTime, "hide ctx", Color.CONTEXT));
    }

}
