package de.unihannover.orderingExperimentAnalysis;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class SessionData {

    private static final int NO_TIME_LIMIT = 99999;

    public static interface Defect {

        public static<T extends Defect> void putDef(
                        Map<String, List<Defect>> ret, String remark, T defectClassification, T... furtherDefects) {
            final List<Defect> defects = new ArrayList<>();
            defects.add(defectClassification);
            defects.addAll(Arrays.asList(furtherDefects));
            ret.put(remark, defects);
        }

        public abstract String getShortString();

        public abstract boolean isDelocalized();

    }

    public enum FP implements Defect {
        FP;

        @Override
        public String getShortString() {
            return this.getClass().getSimpleName();
        }

        @Override
        public boolean isDelocalized() {
            return false;
        }
    }

    private static final Pattern WMSPAN_TRIAL_PATTERN = Pattern.compile("trial result,([0-9]+),([0-9]+),([0-9]+)");
    private static final Pattern WMSPAN_TRIAL_PATTERN_EXTRA_ONE = Pattern.compile("trial result extra,([0-9]+),([0-9]+),(10)");
    private static final Pattern WMSPAN_TRIAL_PATTERN_EXTRA_TWO = Pattern.compile("trial result extra,([0-9]+),([0-9]+),(12)");
    

    private final Set<String> DUPLICATE_BLACKLIST = new HashSet<>(Arrays.asList(
                    "239824586668", "239824586671",
                    "239373198651", "239373198683"));



    private final List<LogRecord> records;
    private final Map<Integer, List<LogRecord>> reviewRecordCache = new HashMap<>();

    public SessionData(List<LogRecord> records) {
        this.records = records;
        Collections.sort(this.records, new Comparator<LogRecord>() {
            @Override
            public int compare(LogRecord o1, LogRecord o2) {
                return Long.compare(o1.getTimestamp(), o2.getTimestamp());
            }
        });
    }

    public String getId() {
        return this.records.get(0).getSessionId();
    }

    public String getShortId() {
        return this.getRestAfterMarker("alpha id is ", null);
    }

    private String getRestAfterMarker(final String marker, String defaultValue) {
        final Optional<LogRecord> value = this.findFirstStartingWith(marker);
        if (!value.isPresent() && defaultValue != null) {
            return defaultValue;
        }
        return value.get().getRest().substring(marker.length());
    }

    private Optional<LogRecord> findFirstStartingWith(String prefix) {
        return this.records.stream().filter((LogRecord l) -> l.getRest().startsWith(prefix)).findFirst();
    }

    public long getTotalDuration() {
        LogRecord lastNonCancelRecord = null;
        for (int i = this.records.size() - 1; i > 0; i--) {
            final LogRecord cur = this.records.get(i);
            final String rest = cur.getRest().trim();
            if (!(rest.equals("experiment cancelled") || rest.equals("change state;cancelled"))) {
                lastNonCancelRecord = cur;
                break;
            }
        }
        return this.findFirstStartingWith("experiment started").get().timeUntil(lastNonCancelRecord);
    }

    public String getVariable(String name) {
    	try {
    		return this.findFirstStartingWith("set variable;" + name + ";").get().getSuffixOfRest(2);
    	}
    	catch(NoSuchElementException e) {
    		return "NA";
    	}
    }

    public String getContextAnswerSmallReview() {
        final Pattern p = Pattern.compile("Small question: ([A-z]*)");
        final Matcher m = this.findMatchInReviewData(0, p);
        return m == null ? "NA" : m.group(1);
    }

    public String getContextAnswerLargeAReview() {
        final Pattern p = Pattern.compile("Task question: ([A-z]*)");
        final Matcher m = this.findMatchInReviewData(1, p);
        if(m == null) {
            final Matcher m2 = this.findMatchInReviewData(2, p);
            return m2 == null ? "NA" : m2.group(1);
        }
        else {
            return m.group(1);
        }
    }

    public String getContextAnswerLargeBReview() {
        final Pattern p = Pattern.compile("Hypersearch question: ([A-z]*)");
        final Matcher m = this.findMatchInReviewData(1, p);
        if(m == null) {
            final Matcher m2 = this.findMatchInReviewData(2, p);
            return m2 == null ? "NA" : m2.group(1);
        }
        else {
            return m.group(1);
        }
    }

    public String getGuidanceApproach() {
        try {
            return this.findFirstStartingWith("guidance approach;").get().getSuffixOfRest(2);
        }
        catch(NoSuchElementException e) {
            return "NA";
        }
    }

    public String getVariable(String name, String defaultValue) {
        final Optional<LogRecord> record = this.findFirstStartingWith("set variable;" + name + ";");
        return record.isPresent() ? record.get().getSuffixOfRest(2) : defaultValue;
    }

    public int getUnderstandingCorrect(int reviewNumber) {
        switch (reviewNumber) {
        case 0:
            return this.getUnderstandingCorrectWarmup();
        case 1:
        case 2:
            return this.getChange(reviewNumber) == Patch.P1 ? this.getUnderstandingCorrectC1() : this.getUnderstandingCorrectC2();
        default:
            throw new AssertionError();
        }
    }

    public int getUnderstandingTotal(int reviewNumber) {
        switch (reviewNumber) {
        case 0:
            return 2;
        case 1:
        case 2:
            return this.getChange(reviewNumber) == Patch.P1 ? 4 : 3;
        default:
            throw new AssertionError();
        }
    }

    public int getUnderstandingCorrectWarmup() {
        return this.questionCorrectWA()
            + this.questionCorrectWB();
    }

    public int questionCorrectWB() {
        return toInt(this.getVariable("methodAdded", "").equals("1"));
    }

    public int questionCorrectWA() {
        return toInt(this.getVariable("idStatement", "").equals("2"));
    }

    public int getUnderstandingCorrectC1() {
        return this.questionCorrectC1A()
            + this.questionCorrectC1B()
            + this.questionCorrectC1C()
            + this.questionCorrectC1D();
    }

    public int questionCorrectC1A() {
        return toInt(this.getVariable("changeDescription", "").equals("0"));
    }

    public int questionCorrectC1B() {
        return toInt(this.getVariable("loadInfo", "").equals("0"));
    }

    public int questionCorrectC1C() {
        return toInt(this.getVariable("deletionTask", "").equals("2"));
    }

    public int questionCorrectC1D() {
        return toInt(this.getVariable("renaming", "").equals("0"));
    }

    public int getUnderstandingCorrectC2() {
        return this.questionCorrectC2A()
            + this.questionCorrectC2B()
            + this.questionCorrectC2C();
    }

    public int questionCorrectC2A() {
        return toInt(this.getVariable("charChange", "").equals("0"));
    }

    public int questionCorrectC2B() {
        return toInt(this.getVariable("nextMatchImpl", "").equals("2"));
    }

    public int questionCorrectC2C() {
        return toInt(this.getVariable("matcherCreation", "").equals("0"));
    }

    private static int toInt(boolean bool) {
        return bool ? 1 : 0;
    }

    private static final class ReviewStateTracer {
        private final boolean includeIdling;
        private boolean reviewActive = true;
        private boolean pauseActive = false;
        private boolean descActive = false;
        private boolean idlingActive = false;
        private long activeTime;
        private LogRecord prev;

        public ReviewStateTracer(LogRecord firstRecord, boolean includeIdling) {
            this.prev = firstRecord;
            this.includeIdling = includeIdling;
        }

        public void handle(LogRecord r) {
            final boolean wasActive = this.isActive();

            final String rest = r.getSuffixOfRest(3);
            if (rest.startsWith("idling started")) {
                this.idlingActive = true;
            } else if (rest.startsWith("idling ended")) {
                this.idlingActive = false;
            } else if (rest.startsWith("pause started")) {
                this.pauseActive = true;
            } else if (rest.startsWith("pause ended")) {
                this.pauseActive = false;
            } else if (rest.startsWith("desc started")) {
                this.descActive = true;
            } else if (rest.startsWith("desc ended")) {
                this.descActive = false;
            } else if (rest.startsWith("review ended")) {
                this.reviewActive = false;
            }

            final boolean willBeActive = this.isActive();

            if (wasActive != willBeActive) {
                if (wasActive) {
                    this.activeTime += this.prev.reviewTimeUntil(r);
                    this.prev = null;
                } else {
                    this.prev = r;
                }
            }
        }

        private boolean isActive() {
            if (this.includeIdling) {
                //include idling := include idle time in sum => don't respect idle in active check
                return this.reviewActive && !this.pauseActive && !this.descActive;
            } else {
                return this.reviewActive && !this.pauseActive && !this.descActive && !this.idlingActive;
            }
        }

        public long timeUntil(LogRecord r) {
            if (this.isActive()) {
                return this.activeTime + this.prev.reviewTimeUntil(r);
            } else {
                return this.activeTime;
            }
        }

    }

    public long getNetTimeForReview(int number, boolean includeIdling) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        if (reviewRecords.isEmpty()) {
            return -1;
        }
        final ReviewStateTracer state = new ReviewStateTracer(reviewRecords.get(0), includeIdling);
        for (final LogRecord r : reviewRecords) {
            state.handle(r);
        }
        return state.activeTime;
    }

    public long getGrossTimeForReview(int number) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        if (reviewRecords.isEmpty()) {
            return -1;
        }
        final LogRecord firstRecord = reviewRecords.get(0);
        final LogRecord lastRecord = reviewRecords.get(reviewRecords.size() - 1);
        return firstRecord.reviewTimeUntil(lastRecord);
    }

    public long getNetTimeUntilFound(int number, Defect defect) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        if (reviewRecords.isEmpty()) {
            return -1;
        }
        final ReviewStateTracer state = new ReviewStateTracer(reviewRecords.get(0), true);
        final RemarkClassifier classifier = new RemarkClassifier(number);
        for (final LogRecord r : reviewRecords) {
            state.handle(r);

            final String rest = r.getSuffixOfRest(3).trim();
            if (rest.startsWith("createReviewRemark;") || rest.startsWith("changeReviewRemark;")) {
                final String remark = rest.substring("createReviewRemark;".length());
                final List<Defect> defects = classifier.getDefects(classifier.toRealFilename(remark));
                if (defects != null && defects.contains(defect)) {
                    return state.timeUntil(r);
                }
            }
        }
        return -1;
    }

    public long getGrossTimeForUnderstanding(int number) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(number);
        final Optional<LogRecord> reviewEnd = reviewRecords.stream().filter((LogRecord r) -> r.getSuffixOfRest(3).startsWith("review ended")).findFirst();
        if (!reviewEnd.isPresent()) {
            return -1;
        }
        final LogRecord last = reviewRecords.get(reviewRecords.size() - 1);
        return reviewEnd.get().reviewTimeUntil(last);
    }

    private List<LogRecord> getReviewRecords(int number) {
        if (this.reviewRecordCache.containsKey(number)) {
            return this.reviewRecordCache.get(number);
        }
        final List<LogRecord> ret = this.records.stream().filter((LogRecord r) -> r.getRest().startsWith("log from review;" + number + ";")).collect(Collectors.toList());
        
        this.reviewRecordCache.put(number, Collections.unmodifiableList(ret));
        return ret;
    }

    public Treatment getTreatment(int reviewNumber) {
        if(reviewNumber == -1) {
            return Treatment.AO;
        }
        if (reviewNumber == 0) {
            return Treatment.AO;
        }
        final String combination = this.getRestAfterMarker("assigned treatment combination ", "NA,NA,NA,NA");
        return shortenTreatment(combination.split(",")[reviewNumber - 1].trim());
    }

    private static Treatment shortenTreatment(String longForm) {
        switch (longForm) {
        case "NA":
            return null;
        case "OPTIMAL_FILES":
            return Treatment.OF;
        case "OPTIMAL_NO_FILES":
            return Treatment.ONF;
        case "WORST_FILES":
            return Treatment.WF;
        case "WORST_NO_FILES":
            return Treatment.WNF;
        case "ALPHABETICAL_ORDER":
            return Treatment.AO;
        default:
            throw new AssertionError("unknown treatment " + longForm);
        }
    }

    public String getTreatmentCombination() {
        final Treatment second = this.getTreatment(1);
        if (second == null) {
            return "NA";
        }
        final Treatment third = this.getTreatment(2);
        return secondIsBetter(second, third) ? (second + "-" + third) : (third + "-" + second);
    }

    public Patch getChange(int reviewNumber) {
        if (reviewNumber == -1) {
            return Patch.TU;
        }
        if (reviewNumber == 0) {
            return Patch.WU;
        }
        final String combination = this.getRestAfterMarker("assigned treatment combination ", "NA,NA,NA,NA");
        return shortenChange(combination.split(",")[reviewNumber + 1].trim());
    }

    private static Patch shortenChange(String longForm) {
        switch (longForm) {
        case "NA":
            return null;
        case "DIFF1":
            return Patch.P1;
        case "DIFF2":
            return Patch.P2;
        default:
            throw new AssertionError("unknown change " + longForm);
        }
    }

    public int getDefectsTotalForReview(int reviewNumber) {
        return this.getAllDefectsForReview(reviewNumber).length;
    }

    public Defect[] getAllDefectsForReview(int reviewNumber) {
        return this.getChange(reviewNumber).getDefects();
    }

    public int getDefectsFoundForReview(int reviewNumber) {
        final Set<Defect> defects = this.classifyDefects(reviewNumber, NO_TIME_LIMIT);
        return defects.size();
    }

    public int getDefectsFoundAfter(int reviewNumber, int minuteTreshold) {
        final Set<Defect> defects = this.classifyDefects(reviewNumber, minuteTreshold);
        return defects.size();
    }

    private final class RemarkClassifier {

        Map<String, List<Defect>> classification;
        final Map<Integer, String> filenameMap;

        public RemarkClassifier(int reviewNumber) {
            if(reviewNumber == -1) {
                this.classification = SessionData.this.getTrueDefects(Patch.TU);
                this.filenameMap = new HashMap<>();
                this.filenameMap.put(1, "1");
            }
            else if (reviewNumber == 0) {
                this.classification = SessionData.this.getTrueDefects(Patch.WU);
                this.filenameMap = new HashMap<>();
                this.filenameMap.put(1, "1");
                this.filenameMap.put(2, "2");
                this.filenameMap.put(3, "3");
            } else {
                this.classification = SessionData.this.getTrueDefects(SessionData.this.getChange(reviewNumber));
                this.filenameMap = SessionData.this.determineFilenameMap(SessionData.this.getTreatment(reviewNumber), SessionData.this.getChange(reviewNumber));
            }
        }

        public String toRealFilename(String remark) {
            String ret = remark;
            for (final Entry<Integer, String> e : this.filenameMap.entrySet()) {
                ret = ret.replace("#" + e.getKey() + "R;", "#" + e.getValue() + "R;");
                ret = ret.replace("#" + e.getKey() + "L;", "#" + e.getValue() + "L;");
            }
            return ret;
        }

        public List<Defect> getDefects(String remarkWithRealFilename) {
            return this.classification.get(remarkWithRealFilename);
        }

    }

    private Set<Defect> classifyDefects(int reviewNumber, int timeTresholdMinutes) {
        final Collection<String> remarks = this.determineReviewRemarks(reviewNumber, timeTresholdMinutes);
        final RemarkClassifier classifier = new RemarkClassifier(reviewNumber);
        final Set<Defect> defects = new LinkedHashSet<>();
        for (final String remark : remarks) {
            final String remarkWithRealFilename = classifier.toRealFilename(remark);
            final List<Defect> classified = classifier.getDefects(remarkWithRealFilename);
            if (classified == null || classified.isEmpty()) {
                throw new RuntimeException("unclassified remark in " + this.getChange(reviewNumber) + ": " + remarkWithRealFilename);
            } else if (!classified.contains(FP.FP)) {
                defects.addAll(classified);
            } else {
                assert classified.size() == 1;
            }
        }
        return defects;
    }

    public Collection<String> getUnclassifiedRemarks() {
        final Set<String> unclassified = new TreeSet<>();
        for (int reviewNumber = 0; reviewNumber < 3; reviewNumber++) {
            final Collection<String> remarks = this.determineReviewRemarks(reviewNumber, NO_TIME_LIMIT);
            if (remarks.isEmpty()) {
                continue;
            }
            final RemarkClassifier classifier = new RemarkClassifier(reviewNumber);
            for (final String remark : remarks) {
                final String remarkWithRealFilename = classifier.toRealFilename(remark);
                final List<Defect> classified = classifier.getDefects(remarkWithRealFilename);
                if (classified == null || classified.isEmpty()) {
                    unclassified.add("in " + this.getChange(reviewNumber) + ": " + remarkWithRealFilename);
                }
            }
        }
        return unclassified;
    }

    private Map<Integer, String> determineFilenameMap(Treatment treatment, Patch reviewTask) {
        int taskCode = 0;
        if(reviewTask.equals(Patch.P1)) {
            taskCode = 1;
        }
        if(reviewTask.equals(Patch.P2)) {
            taskCode = 2;
        }
        final String rest = this.findFirstStartingWith("hunkMap;" + treatment.getLongForm() + ";c"+taskCode).get().getRest();
        System.out.println("test review rest "+rest);
        final String map = rest.substring(rest.indexOf('[') + 1, rest.indexOf(']'));
        final Map<Integer, String> ret = new TreeMap<>();
        for (final String part : map.split(",")) {
            final String[] kv = part.trim().split("=");
            ret.put(Integer.parseInt(kv[0]), kv[1]);
        }
        return ret;
    }

    private Collection<String> determineReviewRemarks(int reviewNumber, int timeTresholdMinutes) {
        final Map<String, String> reviewRemarks = new TreeMap<>();
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        if (reviewRecords.isEmpty()) {
            return Collections.emptySet();
        }
        final ReviewStateTracer stateTracer = new ReviewStateTracer(reviewRecords.get(0), true);
        for (final LogRecord r : reviewRecords) {
            stateTracer.handle(r);
            if (stateTracer.timeUntil(r) > timeTresholdMinutes * 60L * 1000) {
                assert timeTresholdMinutes < NO_TIME_LIMIT;
                break;
            }
            final String rest = r.getSuffixOfRest(3).trim();
            if (rest.startsWith("createReviewRemark;")) {
                reviewRemarks.put(this.getRemarkKey(rest), rest.substring("createReviewRemark;".length()));
            } else if (rest.startsWith("changeReviewRemark;")) {
                reviewRemarks.put(this.getRemarkKey(rest), rest.substring("changeReviewRemark;".length()));
            } else if (rest.startsWith("deleteReviewRemark;")) {
                reviewRemarks.remove(this.getRemarkKey(rest));
            }
        }
        return reviewRemarks.values();
    }

    private String getRemarkKey(String rest) {
        final String[] split = rest.split(";");
        return split[1] + ";" + split[2];
    }

    public boolean isReviewsIncomplete() {
        return !this.records.stream().filter((LogRecord l) -> l.getRest().startsWith("log from review;2;")).findFirst().isPresent();
    }

    public boolean isJeditIncomplete() {
        return !this.records.stream().filter((LogRecord l) -> l.getRest().startsWith("set variable;jedit;")).findFirst().isPresent();
    }

    public String interpretComparisonUnderstanding() {
        if (!(this.isRealEffort(1) && this.isRealEffort(2))) {
            return "NA";
        }
        final String rawValue = this.getVariable("comparisonUnderstanding");
        if (rawValue.equals("same")) {
            return rawValue;
        } else {
            final boolean secondActualBetter = rawValue.equals("second");
            if (secondActualBetter == this.isSecondExpectedBetter()) {
                return "expected";
            } else {
                return "inverse";
            }
        }
    }

    public String getComparisonUnderstanding() {
        if (!(this.isRealEffort(1) && this.isRealEffort(2))) {
            return "NA";
        }
        final String rawValue = this.getVariable("comparisonUnderstanding");
        return rawValue;
    }

    public String interpretComparisonComplicated() {
        if (!(this.isRealEffort(1) && this.isRealEffort(2))) {
            return "NA";
        }
        final String rawValue = this.getVariable("comparisonComplicated");
        if (rawValue.equals("same")) {
            return rawValue;
        } else {
            final boolean secondActualBetter = !rawValue.equals("second");
            if (secondActualBetter == this.isSecondExpectedBetter()) {
                return "expected";
            } else {
                return "inverse";
            }
        }
    }

    public String getComparisonComplicated() {
        if (!(this.isRealEffort(1) && this.isRealEffort(2))) {
            return "NA";
        }
        final String rawValue = this.getVariable("comparisonComplicated");
        return rawValue;
    }

    private boolean isSecondExpectedBetter() {
        final Treatment second = this.getTreatment(1);
        final Treatment third = this.getTreatment(2);
        return secondIsBetter(second, third);
    }

    private static boolean secondIsBetter(final Treatment second, final Treatment third) {
        return second.equals(Treatment.OF)
            || second.equals(Treatment.ONF)
            || (second.equals(Treatment.WF) && third.equals(Treatment.WNF));
    }

    public int getWMSpanScore() {
        final List<LogRecord> wmspanRecords = this.getWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        int skips = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                System.out.println("HERE WM  "+ m.group(3));
                if (size < 3) {
                    //skip warmup
                    skips++;
                    continue;
                }
                final int lettersCorrect = Integer.parseInt(m.group(1));
                sum += lettersCorrect;
            }
        }
        if (skips != 2) {
            throw new AssertionError("something is wrong with wmspan in " + this.getId() + ", skipped " + skips);
        }
        if (sum <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        return sum;
    }
    
    public int getSecondWMSpanScore() {
        final List<LogRecord> wmspanRecords = this.getSecondWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        int skips = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                if (size < 3) {
                    //skip warmup
                    skips++;
                    continue;
                }
                final int lettersCorrect = Integer.parseInt(m.group(1));
                sum += lettersCorrect;
            }
        }
        if (skips != 2) {
            throw new AssertionError("something is wrong with wmspan in " + this.getId() + ", skipped " + skips);
        }
        if (sum <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        return sum;
    }

    public int getWMSpanMathErrors() {
        if (this.getWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.getWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                if (size < 3) {
                    //skip warmup
                    continue;
                }
                final int mathCorrect = Integer.parseInt(m.group(2));
                sum += size - mathCorrect;
            }
        }
        return sum;
    }
    
    public int getSecondWMSpanMathErrors() {
        if (this.getSecondWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.getSecondWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                if (size < 3) {
                    //skip warmup
                    continue;
                }
                final int mathCorrect = Integer.parseInt(m.group(2));
                sum += size - mathCorrect;
            }
        }
        return sum;
    }

    public long getWMSpanGrossTime() {
        if (this.getWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.getWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        final LogRecord first = wmspanRecords.get(0);
        final LogRecord last = wmspanRecords.get(wmspanRecords.size() - 1);
        return first.wmspanTimeUntil(last);
    }
    
    public long getSecondWMSpanGrossTime() {
        if (this.getSecondWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.getSecondWMSpanRecords();
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        final LogRecord first = wmspanRecords.get(0);
        final LogRecord last = wmspanRecords.get(wmspanRecords.size() - 1);
        return first.wmspanTimeUntil(last);
    }

    private List<LogRecord> getWMSpanRecords() {
        final List<LogRecord> list = this.records.stream().filter((LogRecord r) -> r.getRest().startsWith("log from wm test;")).collect(Collectors.toList());
        //in one case the data is duplicated in the log (most likely, the user pressed back in the browser), discard it
        final List<LogRecord> filtered = new ArrayList<>();
        boolean hadStart = false;
        for (final LogRecord r : list) {
            if (r.getSuffixOfRest(2).trim().equals("operation span initialized")) {
                if (hadStart) {
                    break;
                } else {
                    hadStart = true;
                }
            }
            filtered.add(r);
        }
        assert list.isEmpty() || hadStart;
        return filtered;
    }
    
    private List<LogRecord> getSecondWMSpanRecords() {
        final List<LogRecord> list = this.records.stream().filter((LogRecord r) -> r.getRest().startsWith("log from second wm test;")).collect(Collectors.toList());
        //in one case the data is duplicated in the log (most likely, the user pressed back in the browser), discard it
        final List<LogRecord> filtered = new ArrayList<>();
        boolean hadStart = false;
        for (final LogRecord r : list) {
            if (r.getSuffixOfRest(2).trim().equals("operation span initialized")) {
                if (hadStart) {
                    break;
                } else {
                    hadStart = true;
                }
            }
            filtered.add(r);
        }
        assert list.isEmpty() || hadStart;
        return filtered;
    }

    /**
     * Checks if there were any revisits of an already opened page and returns these
     * states/pages. Consequently, an empty list is returned if everything is OK.
     */
    public List<String> determinePageReloads() {
        final Set<String> visited = new HashSet<>();
        final List<String> ret = new ArrayList<>();
        for (final LogRecord record : this.records) {
            if (record.getRest().startsWith("change state;")) {
                final String state = record.getSuffixOfRest(1);
                if (visited.contains(state)) {
                    ret.add(state);
                } else {
                    visited.add(state);
                }
            }
        }
        return ret;
    }

    private Map<String, List<Defect>> getTrueDefects(Patch changeId) {
        switch (changeId) {
        case TU:
            return TU.getTrueDefects();
        case WU:
            return WU.getTrueDefects();
        case P1:
            return P1.getTrueDefects();
        case P2:
            return P2.getTrueDefects();
        default:
            throw new AssertionError("unknown id " + changeId);
        }
    }

    public String getLastState() {
        final ListIterator<LogRecord> iter = this.records.listIterator(this.records.size());
        while (iter.hasPrevious()) {
            final LogRecord r = iter.previous();
            if (r.getRest().startsWith("change state;")) {
                return r.getRestField(1);
            }
        }
        throw new AssertionError();
    }

    /**
     * Return true if it seems the participant did a real effort on the review with the given number.
     * It is not a real effort if there are neither remarks the duration is very short.
     * We don't check for the understanding questions here.
     */
    public boolean isRealEffort(int reviewNumber) {
        if (reviewNumber == -1) {
            return true; //In the tutorial, we skip the effort check
        }
        if (reviewNumber == 2 && this.getId().equals("239824586695")) {
            //this participant wrote a remark that he will not continue reviewing
            return false;
        }
        final long duration = this.getNetTimeForReview(reviewNumber, true);
        if (duration >= 5L * 60 * 1000) {
            return true;
        }
        final Collection<String> remarks = this.determineReviewRemarks(reviewNumber, NO_TIME_LIMIT);
        if (!remarks.isEmpty()) {
            return true;
        }
        return false;
    }

    public int foundDefect(int reviewNumber, Defect d) {
        if (!this.isRealEffort(reviewNumber)) {
            return -1;
        }
        return this.classifyDefects(reviewNumber, NO_TIME_LIMIT).contains(d) ? 1 : 0;
    }

    public int getReviewNumberWith(Patch diff) {
        if (this.getChange(1) == diff) {
            return 1;
        } else if (this.getChange(2) == diff) {
            return 2;
        } else {
            assert !this.isRealEffort(1);
            return 1;
        }
    }

    public int getReviewNumberWith(Defect d) {
        if (d instanceof WU) {
            return 0;
        } else if (d instanceof P1) {
            return this.getReviewNumberWith(Patch.P1);
        } else if (d instanceof P2) {
            return this.getReviewNumberWith(Patch.P2);
        } else {
            throw new AssertionError(d);
        }
    }

    public int getViewWidth() {
        final Pattern p = Pattern.compile("window size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? -1 : Integer.parseInt(m.group(1));
    }

    private Matcher findMatchInReviewData(int reviewNumber, Pattern pattern) {
        final List<LogRecord> review = this.getReviewRecords(reviewNumber);
        for (final LogRecord r : review) {
            final String suffix = r.getSuffixOfRest(3).trim();
            final Matcher m = pattern.matcher(suffix);
            if (m.matches()) {
                return m;
            }
        }
        return null;
    }

    public int getViewHeight() {
        final Pattern p = Pattern.compile("window size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? -1 : Integer.parseInt(m.group(2));
    }

    public int getScreenWidth() {
        final Pattern p = Pattern.compile("screen size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? -1 : Integer.parseInt(m.group(1));
    }

    public int getScreenHeight() {
        final Pattern p = Pattern.compile("screen size ([0-9]+) x ([0-9]+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? -1 : Integer.parseInt(m.group(2));
    }

    public String getBrowser() {
        final Pattern p = Pattern.compile("view loaded,user agent=(.+)");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? "NA" : this.translateBrowser(m.group(1));
    }

    private String translateBrowser(String userAgent) {
        if (userAgent.contains("OPR/")) {
            return "Opera";
        } else if (userAgent.contains("Trident/")) {
            return "IE";
        } else if (userAgent.contains("Chrome/")) {
            return "Chrome";
        } else if (userAgent.contains("Safari/")) {
            return "Safari";
        } else if (userAgent.contains("Firefox/")) {
            return "Firefox";
        }
        throw new AssertionError("unknown user agent " + userAgent);
    }

    public String getExternalIpAtStart() {
        final String record = this.getRestAfterMarker("request for ", null);
        final int index = record.indexOf(" from ");
        if (index < 0) {
            return "NA";
        }
        return record.substring(index + 6).trim();
    }

    public String getInternalIp() {
        return this.getVariable("clientSideIpHash", "NA");
    }

    public boolean isPrize() {
        return Long.parseLong(this.getId()) > 238915110303L;
    }

    public boolean isControlledSetting() {
        final String extIp = this.getExternalIpAtStart();
        return extIp.equals("1A6CB90A4B8DBACECD2DD4F5BEBE6095")
            || (extIp.equals("NA") && !this.isPrize() && this.getBrowser().equals("IE"))
            || this.getId().equals("237967248580");
    }

    public String getPreviousSessionR1FromCookie() {
        final Pattern p = Pattern.compile("cookie;lastReview0=[0-9]+; lastReview1=([0-9]+).*");
        final Matcher m = this.findMatchInReviewData(1, p);
        return m == null ? "NA" : m.group(1);
    }

    public boolean isDuplicateParticipation() {
        return this.DUPLICATE_BLACKLIST.contains(this.getId());
    }

    public List<String> getExceptions() {
        return this.records.stream()
                        .filter((LogRecord l) -> l.getRest().startsWith("exception;"))
                        .map((LogRecord l) -> l.getSuffixOfRest(1))
                        .collect(Collectors.toList());
    }

    public long getScrollAmount(int reviewNumber) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        long currentY = 0;
        long scrollSum = 0;
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3);
            if (key.equals("scroll")) {
                final String y = r.getRestField(5).trim();
                final long nextY = Math.round(Double.parseDouble(y));
                scrollSum += Math.abs(currentY - nextY);
                currentY = nextY;
            }
        }
        return scrollSum;
    }

    public long getPauseCount(int reviewNumber) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        int ret = 0;
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3).trim();
            if (key.equals("pause started")) {
                ret++;
            }
        }
        return ret;
    }

    public long getDescriptionCount(int reviewNumber) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        int ret = 0;
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3).trim();
            if (key.equals("desc started")) {
                ret++;
            }
        }
        return ret;
    }

    public NavigationTrace extractNavigationTrace(int reviewNumber) {
        final List<LogRecord> reviewRecords = this.getReviewRecords(reviewNumber);
        final NavigationTrace ret = new NavigationTrace(
                        this.getId(),
                        this.getChange(reviewNumber),
                        this.getTreatment(reviewNumber),
                        reviewNumber,
                        reviewRecords.get(0).getClientTimeReview(),
                        this.getViewHeight());
        final RemarkClassifier classifier = new RemarkClassifier(reviewNumber);
        for (final LogRecord r : reviewRecords) {
            final String key = r.getRestField(3).trim();
            if (key.equals("scroll")) {
                final String y = r.getRestField(5).trim();
                final long yPos = Math.round(Double.parseDouble(y));
                ret.addScrollTo(r.getClientTimeReview(), yPos);
            } else if (key.startsWith("review ended")) {
                ret.addReviewEnded(r.getClientTimeReview());
            } else if (key.equals("mdo")) {
                final String data = r.getRestField(8);
                ret.addMouseClick(r.getClientTimeReview(), data.charAt(0), data.charAt(1) == 'T', data.charAt(2) == 'T', data.charAt(3) == 'T', data.charAt(4) == 'T');
            } else if (key.equals("key")) {
                final String flags = r.getRestField(5);
                ret.addKeyPress(r.getClientTimeReview(), Integer.parseInt(r.getRestField(4)), flags.charAt(0) == 'T', flags.charAt(1) == 'T', flags.charAt(2) == 'T');
            } else if (key.equals("idling started")) {
                ret.addStateChange(r.getClientTimeReview(), true, "idling");
            } else if (key.equals("idling ended")) {
                ret.addStateChange(r.getClientTimeReview(), false, "idling");
            } else if (key.equals("pause started")) {
                ret.addStateChange(r.getClientTimeReview(), true, "pause");
            } else if (key.equals("pause ended")) {
                ret.addStateChange(r.getClientTimeReview(), false, "pause");
            } else if (key.equals("desc started")) {
                ret.addStateChange(r.getClientTimeReview(), true, "descr");
            } else if (key.equals("desc ended")) {
                ret.addStateChange(r.getClientTimeReview(), false, "descr");
            } else if (key.equals("createReviewRemark")) {
                final String remark = classifier.toRealFilename(r.getSuffixOfRest(4).trim());
                final List<Defect> defects = classifier.getDefects(remark);
                if (defects == null || defects.isEmpty()) {
                    throw new RuntimeException("unclassified remark in " + this.getChange(reviewNumber) + ": " + remark);
                }
                for (final Defect d : defects) {
                    ret.addRemark(r.getClientTimeReview(), this.getRemarkKey(";" + remark), d);
                }
            } else if (key.equals("changeReviewRemark")) {
                final String remark = classifier.toRealFilename(r.getSuffixOfRest(4).trim());
                final List<Defect> defects = classifier.getDefects(remark);
                if (defects == null || defects.isEmpty()) {
                    throw new RuntimeException("unclassified remark in " + this.getChange(reviewNumber) + ": " + remark);
                }
                for (final Defect d : defects) {
                    ret.changeRemark(r.getClientTimeReview(), this.getRemarkKey(";" + remark), d);
                }
            } else if (key.equals("deleteReviewRemark")) {
                final String pos = this.getRemarkKey(";" + classifier.toRealFilename(r.getSuffixOfRest(4)));
                ret.deleteRemark(r.getClientTimeReview(), pos);
            } else if (key.equals("activateContext")) {
                ret.showContext(r.getClientTimeReview());
            } else if (key.equals("deactivateContext")) {
                ret.hideContext(r.getClientTimeReview());
            }
        }
        return ret;
    }
    
    /*
     * Methods for extra WM tests parts
     */
    
    private List<LogRecord> assignSpanRecords(int wmTestNumber){
    	if(wmTestNumber == 1) {
    		return this.getWMSpanRecords();
    	}
    	else {
    		return this.getSecondWMSpanRecords();
    	}
    }
    
    
    public int getWMSpanScoreExtraTenLetters(int wmTestNumber) {
    	final List<LogRecord> wmspanRecords = this.assignSpanRecords(wmTestNumber);  	
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        int skips = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN_EXTRA_ONE.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                System.out.println(m.group(3));
                if (size < 3) {
                    //skip warmup
                    skips++;
                    continue;
                }
                final int lettersCorrect = Integer.parseInt(m.group(1));
                sum += lettersCorrect;
            }
        }
        //if (skips != 2) {
            //throw new AssertionError("something is wrong with wmspan in " + this.getId() + ", skipped " + skips);
       // }
        if (sum <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        return sum;
    }
    
    public int getWMSpanMathErrorsExtraTenLetters(int wmTestNumber) {
        if (this.getWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.assignSpanRecords(wmTestNumber);
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN_EXTRA_ONE.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                if (size < 3) {
                    //skip warmup
                    continue;
                }
                final int mathCorrect = Integer.parseInt(m.group(2));
                sum += size - mathCorrect;
            }
        }
        return sum;
    }
    
    public int getWMSpanScoreExtraTwelveLetters(int wmTestNumber) {
        final List<LogRecord> wmspanRecords = this.assignSpanRecords(wmTestNumber);
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        int skips = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN_EXTRA_TWO.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                System.out.println(m.group(3));
                if (size < 3) {
                    //skip warmup
                    skips++;
                    continue;
                }
                final int lettersCorrect = Integer.parseInt(m.group(1));
                sum += lettersCorrect;
            }
        }
        //if (skips != 2) {
            //throw new AssertionError("something is wrong with wmspan in " + this.getId() + ", skipped " + skips);
       // }
        if (sum <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        return sum;
    }
    
    public int getWMSpanMathErrorsExtraTwelveLetters(int wmTestNumber) {
        if (this.getWMSpanScore() <= 0) {
            //when somebody got nothing right, he probably had a problem with the widget
            return -1;
        }
        final List<LogRecord> wmspanRecords = this.assignSpanRecords(wmTestNumber);
        if (wmspanRecords.isEmpty()) {
            return -1;
        }
        int sum = 0;
        for (final LogRecord r : wmspanRecords) {
            final Matcher m = WMSPAN_TRIAL_PATTERN_EXTRA_TWO.matcher(r.getSuffixOfRest(2).trim());
            if (m.matches()) {
                final int size = Integer.parseInt(m.group(3));
                if (size < 3) {
                    //skip warmup
                    continue;
                }
                final int mathCorrect = Integer.parseInt(m.group(2));
                sum += size - mathCorrect;
            }
        }
        return sum;
    }

    /*
    New methods for checklist
     */

    public String findChecklistElement(String checklistElement, int reviewNumber) {

        if(!getGuidanceApproach().equals("CHECKLIST")) {
            return "NA";
        }

        String patternWithDefects = "Checklist " + checklistElement + " with issues";
        String patternOK = "Checklist " + checklistElement + " no issues";
        final Pattern pIssues = Pattern.compile(patternWithDefects);
        final Pattern pOK = Pattern.compile(patternOK);

        final Matcher m = this.findMatchInReviewData(reviewNumber, pIssues);
        final Matcher m2 = this.findMatchInReviewData(reviewNumber, pOK);

        if(m == null) {
            if(m2 == null){
                return "NA";
            }
            else {
                return "YES";
            }
        }
        else {
            if(m2 == null) {
                    return "NO";
            }
            else {
                return checkLastElementSelected(reviewNumber, pIssues, pOK);
            }
        }
    }

    private String checkLastElementSelected(int reviewNumber, Pattern patternWithIssues, Pattern patternOK) {

        long issuesTimeTag = 0;
        long okTimeTag = 0;

        int index = 0;
        int okIndex = 0;
        int issuesIndex = 0;

        final List<LogRecord> review = this.getReviewRecords(reviewNumber);
        for (final LogRecord r : review) {
            final String suffix = r.getSuffixOfRest(3).trim();
            final Matcher mIssues = patternWithIssues.matcher(suffix);
            final Matcher mOk = patternOK.matcher(suffix);
            while (mIssues.find()) {
                issuesTimeTag = r.getTimestamp();
                issuesIndex = index;
            }
            while (mOk.find()) {
                okTimeTag = r.getTimestamp();
                okIndex = index;
            }
            index++;
        }

        if(issuesTimeTag > okTimeTag) {
            return "NO";
        }
        else if (okTimeTag > issuesTimeTag) {
            return "YES";
        }
        else {
            if(issuesIndex > okIndex){
                return "NO";
            }
            else if(okIndex > issuesIndex) {
                return "YES";
            }
        }
        return "ERROR";
    }


    /*
    Retrieves strategy answers
     */
    public String findStrategyElement(int strategyStep, int classIndex, int chunkIndex, int reviewNumber) {

        String patternYes = "Strategy step;" + strategyStep + "class;" + classIndex + "index;" + chunkIndex + "answer;" + "yes";
        String patternNo = "Strategy step;" + strategyStep + "class;" + classIndex + "index;" + chunkIndex + "answer;" + "no";
        Pattern pYes = Pattern.compile(patternYes);
        Pattern pNo = Pattern.compile(patternNo);

        System.out.println(patternYes);
        System.out.println(patternNo);

        Matcher mYes = this.findMatchInReviewData(reviewNumber, pYes);
        Matcher mNo = this.findMatchInReviewData(reviewNumber, pNo);

        if(mYes == null) {
            if(mNo == null){
                return "NA";
            }
            else {
                return "No";
            }
        }
        else {
            if(mNo == null) {
                return "Yes";
            }
            else {
                return "Yes - second try"; //If both mYes and mNo are not null then an initial No was changed into a Yes after the warning.
            }
        }
    }

}
