package de.unihannover.orderingExperimentUi;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

public class Treatments {

    public static final class TreatmentCombination {
        private final Treatment t1;
        private final Treatment t2;
        private final Diff d1;
        private final Diff d2;

        public TreatmentCombination(Treatment t1, Diff d1, Diff d2) {
            this.t1 = t1;
            this.t2 = t1;
            this.d1 = d1;
            this.d2 = d2;
        }

        @Override
        public boolean equals(Object o) {
            if (!(o instanceof TreatmentCombination)) {
                return false;
            }
            final TreatmentCombination tc = (TreatmentCombination) o;
            return this.t1.equals(tc.t1)
                && this.t2.equals(tc.t2)
                && this.d1.equals(tc.d1)
                && this.d2.equals(tc.d2);
        }

        @Override
        public int hashCode() {
            return this.t1.hashCode() + this.d1.hashCode();
        }

        @Override
        public String toString() {
            return this.t1 + ", " + this.t2 + ", " + this.d1 + ", " + this.d2;
        }

        public Diff getD1() {
            return this.d1;
        }

        public Diff getD2() {
            return this.d2;
        }

        public Treatment getT1() {
            return this.t1;
        }

        public Treatment getT2() {
            return this.t2;
        }
    }

    public static final class TreatmentUsageData {
        private int count;
        private int reviewScoreSum;

        public TreatmentUsageData() {
            this(0, 0);
        }

        public TreatmentUsageData(int count, int reviewScoreSum) {
            this.count = count;
            this.reviewScoreSum = reviewScoreSum;
        }


        public void remove(int testReviewScore) {
            this.count--;
            this.reviewScoreSum -= testReviewScore;
        }

        public void adjust(int testReviewScore) {
            this.count++;
            this.reviewScoreSum += testReviewScore;
        }

        public void adjust(TreatmentUsageData value) {
            this.count += value.count;
            this.reviewScoreSum += value.reviewScoreSum;
        }

        public double getReviewScoreAvg() {
            return ((double) this.reviewScoreSum) / this.count;
        }

        public TreatmentUsageData copy() {
            final TreatmentUsageData ret = new TreatmentUsageData();
            ret.adjust(this);
            return ret;
        }

        @Override
        public String toString() {
            return "count=" + this.count + ", reviewScoreSum=" + this.reviewScoreSum;
        }
    }

    private static final double EPSILON = 0.0000001;

    private final Map<TreatmentCombination, TreatmentUsageData> map = new LinkedHashMap<>();
    private final Random random;

    public Treatments(Random r) {
        this.random = r;
        //fill the map with the combinations we are interested in (i.e. where the theory makes a prediction)
        this.addPermutationsToMap(Treatment.ALPHABETICAL_ORDER);
        
    }

    private void addPermutationsToMap(Treatment t1) {
        this.map.put(new TreatmentCombination(t1, Diff.DIFF1, Diff.DIFF2), new TreatmentUsageData());
        this.map.put(new TreatmentCombination(t1, Diff.DIFF2, Diff.DIFF1), new TreatmentUsageData());
    }

    public synchronized TreatmentCombination getNextBestTreatment(
                    int testReviewScore) {
        final List<TreatmentCombination> minCombinations = this.determineCombinationsWithMinCount();

        List<TreatmentCombination> toSampleFrom = minCombinations;

        final int index = this.random.nextInt(toSampleFrom.size());
        final TreatmentCombination ret = toSampleFrom.get(index);
        this.map.get(ret).adjust(testReviewScore);
        return ret;
    }

    private List<TreatmentCombination> determineCombinationsWithMinCount() {
        final int minCount = this.determineMinCountOfTreatmentCombination(this.map.keySet());

        final List<TreatmentCombination> minCombinations = new ArrayList<>();
        for (final TreatmentCombination key : this.map.keySet()) {
            if (this.map.get(key).count == minCount) {
                minCombinations.add(key);
            }
        }
        return minCombinations;
    }

    private int determineMinCountOfTreatmentCombination(Collection<TreatmentCombination> toCheck) {
        int minCount = Integer.MAX_VALUE;
        for (final TreatmentCombination key : toCheck) {
            minCount = Math.min(minCount, this.map.get(key).count);
        }
        return minCount;
    }

    public synchronized Map<TreatmentCombination, TreatmentUsageData> getEntries() {
        final Map<TreatmentCombination, TreatmentUsageData> ret = new LinkedHashMap<>();
        for (final Entry<TreatmentCombination, TreatmentUsageData> e : this.map.entrySet()) {
            ret.put(e.getKey(), e.getValue().copy());
        }
        return ret;
    }

    public synchronized void remove(TreatmentCombination treatmentCombination, int testReviewScore) {
        this.map.get(treatmentCombination).remove(testReviewScore);
    }

    public synchronized void add(TreatmentCombination treatmentCombination, int count, int reviewScoreSum) {
        this.map.get(treatmentCombination).adjust(new TreatmentUsageData(count, reviewScoreSum));
    }

}
